%% MODEL ORDER REDUCTION BY PARAMETER-VARYING OBLIQUE PROJECTION
% Computes a reduced order model of the nonlinear
% spring-mass-damper system for 50 blocks using a
% parameter-varying oblique projection
clearvars
%% Load nonlinear mass-spring-damper model from [AnnoniSeiler IJRNC 2015]

addpath('.\example_AnnoniSeiler')
M = 50;         % # of masses
m = 1;          % Mass of individual blocks, kg
k1 = 0.5;       % Linear spring constant, N/m
k2 = 1;         % Cubic spring constant, N/m^3
b1 = 1;         % Linear damping constant, N/(m/s)
b2 = 0;         % Cubic damping constant, N/(m/s)^3
nx =2*M;

% Select Parameter Grid
rhogrid = 0:1:2;
ngrid = numel(rhogrid);

for kk=1:ngrid   
    xtrim(:,kk) = ComputeMSDTrim(rhogrid(kk),M,m,k1,k2,b1,b2);
    Glin = LinearizedMSD(xtrim(:,kk),m,k1,k2,b1);
    sys(:,:,kk) = Glin;
end

%% Run order Reduction 
% requires about 0.1 seconds runtime

% Select reduced order 
nz = 4;

tic
[ sysr, V, W, info ] = projred(sys, nz);
comp_time = toc

%% Make Bode plot
b = bodeoptions(); b.PhaseMatching='on';
figure(21)
bode(sys,sysr,'r--',b)
legend(['Full Order (' int2str(size(sys.a,1)) ')'], ['LPV Reduction (' int2str(size(sysr.a,1)) ')'])

%% Simulations: Compare Nonlinear and Linear Parameter-Varying Results

% Specify sinusoidal parameter trajectory rho(t)
Tf = 50;
dt = 0.1;
tin = 0:dt:Tf;
Nt=numel(tin);

amp = -1;
bias = 1;
freq = 0.5;
rhoin = amp*cos(freq*tin)+bias;
F = [tin(:) rhoin(:)];

for ii=1:ngrid
    % Grab current grid point
    rhoi = rhogrid(ii);
    % Compute equlibrium condition
    xtrim(:,ii) = ComputeMSDTrim(rhoi,M,m,k1,k2,b1,b2);
end

% Input Force, N
u = [tin(:) zeros(Nt,1)];
u(tin>=25,2) = 0.5;

% Function handle to compute the system derivative
xdotfh = @(t,x) NonlinearMSD(t,x,m,k1,k2,b1,b2,F,u);

% Initial Conditions: Use trim condition associated with initial rho
x0 = LPVLinearInterp(rhogrid,xtrim,rhoin(1));
W0 = LPVLinearInterp(rhogrid,W,rhoin(1));  
% Simulate Nonlinear System
[tnl,xnl]=ode45(xdotfh,tin,x0);
ynl = xnl(:,M);

% Linear Simulation with simple Euler discretization
delz = zeros(nz,Nt);
dely = zeros(1,Nt);
zbar = zeros(nz,Nt);
zbar(:,1) = W0'*x0;
xbar = zeros(nx,Nt);
xbar(:,1) = x0;
ybar = zeros(1,Nt);
ybar(:,1) = x0(M);

for k=1:(Nt-1)
    % Parameter Varying projector at time k and k+1
    Wk = LPVLinearInterp(rhogrid,W,rhoin(k));
    Wkp1 = LPVLinearInterp(rhogrid,W,rhoin(k+1));
    
    % Trim Condition at time k and k+1
    xbark = LPVLinearInterp(rhogrid,xtrim,rhoin(k));
    xbarkp1 = LPVLinearInterp(rhogrid,xtrim,rhoin(k+1));
    
    zbar(:,k+1) = Wkp1'*xbarkp1;
    ybar(:,k+1) = xbarkp1(M);
    delu = u(k,2);
    
    % Linearized state matrix at time k
    Ak = LPVLinearInterp(rhogrid,sysr.A,rhoin(k));
    Bk = LPVLinearInterp(rhogrid,sysr.B,rhoin(k));
    Ck = LPVLinearInterp(rhogrid,sysr.C,rhoin(k));
    Dk = LPVLinearInterp(rhogrid,sysr.D,rhoin(k));
    
    % Update State
    zdot = Ak*delz(:,k) + Bk*delu - ( zbar(:,k+1)-zbar(:,k) )/dt;
    delz(:,k+1) = delz(:,k)+zdot*dt ;
    
    % Update Output
    dely(:,k+1) = Ck*delz(:,k) + Dk*delu;
end
ylpv = dely + ybar;

% Plot:  Output (Block M position) vs time
figure(22)
plot(tnl,ynl,'b',tin,ylpv,'r-.',tin,ybar,'g:');
ylim([min(ybar) 1.1*max(ybar)]);
ylabel('Block M Position y, m');
xlabel('Time, sec');
legend('Nonlin',['LPV (' int2str(size(sysr.a,1)) ' states)'],'Trim','Location','Best');
grid on;
if exist('garyfyFigure','file')
    garyfyFigure;
end

OutputMSE = sum((ynl-ylpv').^2)/size(ynl,1)
    
%% CALCULATE UPPER BOUND FOR THE ERROR SSYSTEM
% requires about 70 minutes runtime
% requires LPVTools toolbox
Wlpv = pmat(W,rgrid(pgrid('rho',[0 1 2])));
syslpv = pss(sys,rgrid(pgrid('rho',[0 1 2])));
Pi = V*Wlpv';

[ny, nu] = size(sys);
[A, B, C] = ssdata(syslpv);
A_err = [A     (eye(nx)-Pi)*A*V;
        zeros(nz,nx)  Wlpv'*A*V];
B_err = [(eye(nx)-Pi)*B;  Wlpv'*B];

C_err = [C,  zeros(ny,nz)];

ERR = ss(A_err,B_err,C_err,0);

tic
[Gamma,X] = lpvnorm(ERR,'L2');
toc

% Alternatively, load precomputed balanced model
% load('error_bound_proj.mat')